from typing import Any, List

from pyruicore.data_type import (
    TYPE_MAPPING,
    ListType,
    UserDefineType,
    analysis_annotation,
)
from pyruicore.field import Field


class ModelMetaClass(type):
    def __new__(mcs, cls_name: str, bases: tuple, attrs: dict):
        __fields_map__: [str, Field] = {}

        copy_bases(__fields_map__, bases)
        init_annotations(__fields_map__, attrs)

        attrs["__fields__"] = tuple(__fields_map__.values())
        attrs["__fields_map__"] = __fields_map__
        attrs["__slots__"] = (
            tuple(attrs.get("__slots__", ()))
            + tuple(__fields_map__.keys())
            + ("__storage__",)
        )

        return type.__new__(mcs, cls_name, bases, attrs)


def copy_bases(__fields_map__, bases):
    """copy all field in base class to current class, ignore repeat field, in-place"""
    for base in bases:
        for field in getattr(base, "__fields__", ()):
            if field.name not in __fields_map__:
                __fields_map__[field.name] = field


def init_annotations(__fields_map__, attrs):
    """init Field from user given annotations, in-place"""
    for field_name, anno_type in attrs.get("__annotations__", {}).items():
        if is_reserve_word(field_name):
            continue
        user_ini_field = attrs.pop(field_name, None)  # user init Field
        __fields_map__[field_name] = init_field(user_ini_field, anno_type)


def init_field(user_ini_field: Field, anno_type: [int, str, float, List, Any]):
    nullable, user_field_type, user_ele_type = parse_annotation(anno_type)
    field_type = init_field_type(user_field_type, user_ele_type)

    field = parse_ugv_field(user_ini_field) or Field()
    field.nullable, field.field_type = nullable, field_type
    return field


def is_reserve_word(word: str):
    """word start with `_` is treated as a reserve word"""
    return word.startswith("_")


def is_model_type(field_type: Any):
    """ check if field_type is user define model """
    return isinstance(field_type, ModelMetaClass)


def init_field_type(user_field_type, user_ele_type):
    """ init `field_type` for `Field`"""
    if is_model_type(user_field_type):
        return init_model_type(user_field_type)

    field_type = TYPE_MAPPING.get(user_field_type)
    if field_type is None:
        raise Exception(f"暂时不支持 {user_field_type} 类型")

    if field_type is ListType:
        return init_list_type(user_ele_type)

    return field_type()  # basic type <IntType,StringType ... > we can just init it


def parse_ugv_field(user_ini_field):
    """
    parse user given Field:
    if user_ini_field is type <BaseType>, do nothing, return it;
    else treat user_ini_field as a default value
    """
    if not isinstance(user_ini_field, Field):
        user_ini_field = Field(default=user_ini_field)
    return user_ini_field


def init_model_type(class_type: Any) -> Any:
    """init field_type of type <UserDefineType>"""
    return UserDefineType(class_=class_type)


def init_list_type(field_ele_type):
    """ init field_type of <ListType> """
    if isinstance(field_ele_type, ModelMetaClass):
        return ListType(element_type=UserDefineType(class_=field_ele_type))

    ele_type = TYPE_MAPPING.get(field_ele_type)
    if ele_type is None:
        raise Exception(f"暂时不支持 {field_ele_type} 类型")
    return ListType(element_type=ele_type())


def parse_annotation(field_type):
    """ get `nullable`, `type`, `ele_type(if is List)` from annotation"""
    try:
        nullable, user_field_type, user_ele_type = analysis_annotation(field_type)
    except Exception as e:
        raise e

    return nullable, user_field_type, user_ele_type



from collections.abc import Iterable
from datetime import datetime
from random import choice, randint, random
from typing import Any, List, Optional, Union

from pyruicore.data_type.util import str_to_datetime


class BaseType:
    def mock(self) -> Any:
        raise NotImplementedError()

    def parse(self, field, value) -> Any:
        raise NotImplementedError()

    def validate(self, value) -> Any:
        raise NotImplementedError()

    def marshal(self, value) -> Any:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.__class__.__name__

    __repr__ = __str__
    __name__ = "base_type"


class IntType(BaseType):
    def mock(self) -> int:
        return randint(0, 10)

    def parse(self, field, value) -> Optional[int]:
        try:
            return int(value) if value not in (None, "") else None
        except TypeError:
            raise TypeError(f"{field} expect <class 'int'>, but get {type(value)}")

    def marshal(self, value) -> Optional[int]:
        return int(value) if value else None

    def validate(self, value) -> None:
        assert value is None or isinstance(
            value, int
        ), f"expect <class 'int'>, but get {type(value)}"

    __name__ = "int"


class FloatType(BaseType):
    def mock(self) -> float:
        return int(random() * 100) / 10

    def parse(self, field, value) -> float:
        try:
            return float(value) if value not in (None, "") else None
        except TypeError:
            raise TypeError(f"{field} expect <class 'float'>, but get {type(value)}")

    def marshal(self, value) -> Optional[float]:
        return float(value) if value else None

    def validate(self, value) -> None:
        assert value is None or isinstance(
            value, float
        ), f"expect <class 'float'>, but get {type(value)}"

    __name__ = "float"


class StringType(BaseType):
    def mock(self) -> str:
        return str(random() * 100 / 10)

    def parse(self, field, value) -> Optional[str]:
        try:
            return str(value) if value not in (None, "") else None
        except TypeError:
            raise TypeError(f"{field} expect <class 'str'>, but get {type(value)}")

    def marshal(self, value) -> Optional[str]:
        return str(value) if value else None

    def validate(self, value) -> None:
        assert value is None or isinstance(
            value, str
        ), f"expect <class 'str'>, but get {type(value)}"

    __name__ = "str"


class BooleanType(BaseType):
    def mock(self) -> bool:
        return choice([True, False])

    def parse(self, field, value) -> bool:
        return True if value else False

    def validate(self, value) -> None:
        assert value is None or isinstance(
            value, bool
        ), f"expect <class 'bool'>, but get {type(value)}"

    def marshal(self, value):
        return True if value else False

    __name__ = "bool"


class DateTimeType(BaseType):
    def mock(self) -> datetime:
        return datetime.fromtimestamp(1000000000 * random())

    def parse(self, field, value) -> datetime:
        if value and value not in {"", "null", "None"}:
            if isinstance(value, datetime):
                return value
            elif isinstance(value, str):
                return str_to_datetime(value)

        raise ValueError(
            f"{field} expect datetime, but get{value},Invalid datetime type"
        )

    def marshal(self, value):
        if isinstance(value, datetime):
            return value.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return None

    def validate(self, value) -> None:
        assert value is None or isinstance(
            value, datetime
        ), f"expect <class 'datetime'>, but get {type(value)}"

    __name__ = "datetime"


class ListType(BaseType):
    """ list 目前仅支持单一类型"""

    def __init__(self, element_type: Union[BaseType, Any]) -> None:
        self.element_type = element_type

    def mock(self) -> List[Any]:
        return [self.element_type.mock() for _ in range(10)]

    def parse(self, field, value) -> List[Any]:
        try:
            return (
                [] if not value else [self.element_type.parse(field, v) for v in value]
            )
        except TypeError as e:
            raise e

    def validate(self, value) -> None:
        assert isinstance(value, Iterable) and not isinstance(
            value, str
        ), f"expect <Iterable> and not <class 'str'>, but get {type(value)}"
        [self.element_type.validate(v) for v in value]  # 基本类型

    def marshal(self, value) -> List[Any]:
        return [self.element_type.marshal(v) for v in value]

    def __str__(self):
        return f"List of <{self.element_type.__name__}>"

    __name__ = "list"



from typing import Any, Dict, Union

from pyruicore.data_type import BaseType


class UserDefineType(BaseType):
    def __init__(self, class_: Any):
        self.class_name = class_.__name__
        self.class_type = class_

    def mock(self):
        raise NotImplementedError()

    def parse(self, field, value: Union[Dict[str, Any]]) -> Any:
        from pyruicore.model.base import BaseModel

        if isinstance(value, BaseModel):
            return value
        elif isinstance(value, dict):
            return self.class_type(**value)
        else:
            raise TypeError(
                f"<{field}> expect: <class '{self.class_name}'>, but get: {type(value)}"
            ) from None
            # todo: raise exception ?

    def validate(self, value: Any) -> None:
        assert isinstance(
            value, self.class_type
        ), f"Expect: {self.class_name} - Actual: {type(value)}"
        self.parse(value)

    def marshal(self, value) -> Any:
        return value.dict()

    @property
    def name(self):
        return self.class_name

    def __str__(self):
        return self.class_name

    __name__ = name



from typing import Any, Callable, Tuple, Union, get_args, get_origin

NoneType = type(None)


def is_origin_type(type_hint):
    return isinstance(type_hint, type)


def parse_union(args):
    """
    Optional[int] ==> (True, <class 'int'>,)
    Union[int,None] ==>  (True, <class 'int'>,)
    """

    nullable, types_ = False, []
    for type_ in args:
        if type_ is NoneType:
            nullable = True
        else:
            types_.append(type_)

    if len(types_) == 1:
        real_type = types_[0]
        if get_args(real_type):
            _, real_type, ele_type = analysis_annotation(real_type)
            return nullable, real_type, ele_type
        else:
            return nullable, real_type, None

    raise Exception(f"不支持嵌套类型{args}")


def check_args(args, error=""):
    if len(args) > 2:
        raise Exception(f"不支持嵌套类型{args}")
    if len(args) == 2 and NoneType not in args:
        raise Exception(f"不支持嵌套类型{args}")
    assert all(is_origin_type(arg) for arg in args), f"{error}: {args}"


def analysis_annotation(type_hint: Any) -> Tuple[bool, Any, Any]:
    """
    根据 typing 获取真实的类型
    返回是否可空，以及真实的类型
    """
    if is_origin_type(type_hint) and not get_args(type_hint):
        return False, type_hint, None  # 原始单一类型

    if type_hint is Callable:  # func
        raise Exception(f"不支持 {Callable} 类型") from None

    origin, args = get_origin(type_hint), get_args(type_hint)

    if is_origin_type(origin):
        check_args(args, error=f"类型 {origin} 下不支持")
        return False, origin, args[0]
    if origin is Union:
        return parse_union(args)

    raise Exception(f"不支持 {type_hint} 类型") from None



import pprint
from typing import Any, Dict

from pyruicore.field import Field
from pyruicore.model.meta import ModelMetaClass


class BaseModel(metaclass=ModelMetaClass):
    __fields__ = ()
    __fields_map__: Dict[str, Field] = {}

    def __init__(self, drop_missing=False, **kwargs):
        class_name = type(self).__name__
        for field_name, field in self.__fields_map__.items():
            input_value = kwargs.get(field_name)
            value = field.get_value(input_value)
            if not drop_missing and not field.nullable and not value:
                raise Exception(
                    f"{type(self)}: field <{field_name}> must be initialized"
                )
            setattr(self, field_name, field.parse(f"{class_name}.{field_name}", value))

    def dict(self, exclude_none=False, exclude_unset=False, exclude=None):
        dct = {}
        for field_name, field in self.__fields_map__.items():
            value = self.get(field_name)
            dct[field_name] = field.field_type.marshal(value)

        return dct

    def get(self, key, default=None) -> Any:
        return getattr(self, key, default)

    def __str__(self):
        return f"<{self.__class__.__name__}>: {pprint.pformat(self.dict(), indent=4)}"

    def __repr__(self):
        return self.__class__.__name__



from datetime import datetime

from pyruicore.data_type.const import DATE_FORMATS


def str_to_datetime(value):
    ret = None
    if value and value not in {"", "null", "None"}:
        for format_ in DATE_FORMATS:
            try:
                ret = datetime.strptime(value, format_)
            except ValueError:
                pass
            else:
                break
        else:
            raise Exception(
                f"[date] [time] [datetime] format must be {DATE_FORMATS}, Actual: [{value}]"
            ) from None
    return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyruicore', 'pyruicore.data_type', 'pyruicore.model']

package_data = \
{'': ['*']}

install_requires = \
['typing>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'pyruicore',
    'version': '0.1.2',
    'description': 'load python dict data to python class',
    'long_description': '# pyruicore\n\n[![Build Status](https://travis-ci.com/RuiCoreSci/pyruicore.svg?branch=master)](https://travis-ci.com/RuiCoreSci/pyruicore) &nbsp; [![Coverage Status](https://coveralls.io/repos/github/RuiCoreSci/pyruicore/badge.svg?branch=master)](https://coveralls.io/github/RuiCoreSci/pyruicore?branch=master) &nbsp; [![codebeat badge](https://codebeat.co/badges/af92f04f-6d5e-4a0a-82c6-53a8bcfb0341)](https://codebeat.co/projects/github-com-ruicoresci-pyruicore-master) &nbsp; ![python3.8](https://img.shields.io/badge/language-python3.8-blue.svg) &nbsp; ![issues](https://img.shields.io/github/issues/RuiCoreSci/pyruicore) ![stars](https://img.shields.io/github/stars/RuiCoreSci/pyruicore) &nbsp; ![license](https://img.shields.io/github/license/RuiCoreSci/pyruicore)\n\n* This package is used to load python dict data to python class.\n\n## Usage\n\n* pip install pyruicore -i https://pypi.org/simple\n\n```py\n\n\nfrom pyruicore import BaseModel, Field\n\n\nclass Department(BaseModel):\n    name: str\n    address: str\n\n\nclass User(BaseModel):\n    age: int = Field(default_factory=lambda: 1)\n    departs: List[Department]\n\n\nuser = User(\n    departs=[\n        {"name": "de1", "address": "address1"},\n        Department(name="2", address="address2"),\n    ]\n)\nuser_dict = user.dict()\n"""\nuser_dict = {\n    "age": 1,\n    "departs": [\n        {"name": "de1", "address": "address1"},\n        {"name": "2", "address": "address2"},\n    ]\n}\n"""\n\n```\n\n##  Maintainers\n\n[@ruicore](https://github.com/ruicore)\n\n## Contributing\n\nPRs are accepted, this is first workout version, may have many bugs, so welcome to point out bugs and fix it.\n\n## License\n\nMIT © 2020 ruicore\n',
    'author': 'ruicore',
    'author_email': 'hrui835@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RuiCoreSci/pyruicore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Callable

from pyruicore.data_type import BaseType


class Field:
    __slots__ = (
        "field_type",
        "mock_func",
        "enum_values",
        "comment",
        "nullable",
        "validate",
        "implicit",
        "default",
        "default_factory",
    )

    def __init__(
        self,
        field_type: BaseType = None,
        mock_func: Callable = None,
        enum_values: tuple = (),
        comment: str = "",
        nullable: bool = True,
        validate: Callable = None,
        implicit: bool = True,
        default: Any = None,
        default_factory: Callable = None,
        **kwargs,
    ):
        self.field_type = field_type
        self.mock_func = mock_func
        self.enum_values = enum_values
        self.comment = comment
        self.nullable = nullable
        self.validate = validate
        self.implicit = implicit
        self.default = default
        self.default_factory = default_factory

    def get_value(self, value):
        if value is None:
            if self.default_factory is not None:
                value = self.default_factory()
            elif self.default is not None:
                value = self.default
            else:
                return None
        return value

    def parse(self, field, value):
        if not self.implicit:
            self.field_type.validate(value)
        return self.field_type.parse(field, value)

    def __str__(self):
        return f"<Field of type : {self.field_type or 'unknown yet'}>"

    __repr__ = __str__



"""
all basic type
"""
from datetime import datetime

from pyruicore.data_type.analysis import analysis_annotation
from pyruicore.data_type.basic import (
    BaseType,
    BooleanType,
    DateTimeType,
    FloatType,
    IntType,
    ListType,
    StringType,
)
from pyruicore.data_type.udf import UserDefineType

TYPE_MAPPING = {
    int: IntType,
    str: StringType,
    float: FloatType,
    list: ListType,
    datetime: DateTimeType,
}

__all__ = [
    "BaseType",
    "IntType",
    "FloatType",
    "StringType",
    "BooleanType",
    "DateTimeType",
    "ListType",
    "TYPE_MAPPING",
    "UserDefineType",
    "analysis_annotation",
]
